﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Shapes;
using System.Linq.Expressions;
using System.Threading;
using System.Windows.Media.Animation;
using System.Windows.Threading;
using System.Windows.Media;

namespace VisualLinq
{
    public abstract class VisualEnumerable<T> : IVisualEnumerable, IEnumerable<T>, IDisposable
    {
        public Button Box { get; private set; }
        public Canvas Canvas { get; private set; }

        readonly object syncLock = new object();

        internal VisualEnumerable(Canvas canvas, Point centre, string text)
        {
            Box = new Button();
            Box.Content = text;
            Box.Measure(new Size(10000, 10000));
            Box.Width = Box.DesiredSize.Width;
            Box.Height = Box.DesiredSize.Height;
            Box.IsEnabled = false;
            Box.Background = Brushes.White;
            Box.Foreground = Brushes.Black;
            Canvas = canvas;
            canvas.Children.Add(Box);
            Canvas.SetLeft(Box, centre.X-Box.Width/2);
            Canvas.SetTop(Box, centre.Y-Box.Height/2);
        }

        public abstract IEnumerator<T> GetEnumerator();

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        public VisualEnumerable<T> Where(Expression<Func<T, bool>> predicate)
        {
            return new VisualWhere<T>(this, predicate);
        }

        public VisualEnumerable<TResult> Select<TResult>(Expression<Func<T, TResult>> projection)
        {
            return new VisualSelect<T,TResult>(this, projection);
        }

        public VisualEnumerable<IGrouping<TKey,T>> GroupBy<TKey>(Expression<Func<T, TKey>> keySelector)
        {
            return new VisualGroupBy<T,T,TKey>(this, keySelector, x=>x, false);
        }

        public VisualEnumerable<IGrouping<TKey, TResult>> GroupBy<TKey,TResult>(Expression<Func<T, TKey>> keySelector, Expression<Func<T,TResult>> elementSelector)
        {
            return new VisualGroupBy<T, TResult, TKey>(this, keySelector, elementSelector, true);
        }

        public VisualEnumerable<TResult> SelectMany<TOther, TResult>(Expression<Func<T, VisualEnumerable<TOther>>> selector,
                                                                     Expression<Func<T, TOther, TResult>> resultSelector)
        {
            return new VisualSelectMany<TResult, T, TOther>(this, selector, resultSelector);
        }

        protected void Invoke(Action action)
        {
            Canvas.Dispatcher.Invoke(DispatcherPriority.Normal, action);
        }

        public IVisualBucket End()
        {
            return new VisualBucket<T>(this);
        }

        protected void Animate(Func<Storyboard> animationCreator)
        {
            lock (syncLock)
            {
                Canvas.Dispatcher.BeginInvoke(DispatcherPriority.Normal, (Action<Func<Storyboard>>)RealAnimate, animationCreator);
                Monitor.Wait(syncLock);
            }
        }

        void RealAnimate(Func<Storyboard> animationCreator)
        {
            Storyboard storyboard = animationCreator();
            storyboard.Completed += CompleteAnimation;
            storyboard.Begin(Canvas);
        }

        void CompleteAnimation(object sender, EventArgs e)
        {
            lock (syncLock)
            {
                Monitor.Pulse(syncLock);
            }
        }

        public void Dispose()
        {
            Canvas.Dispatcher.Invoke(DispatcherPriority.Normal, (Action)ThreadedDispose);
        }

        internal virtual void ThreadedDispose()
        {
            Canvas.Children.Remove(Box);
        }
    }
}
